#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

import json
import logging
import os
from secrets import token_hex
from typing import Dict, List

from flask import Flask, Response, request
from matplotlib.patches import CirclePolygon
from osgeo import gdal

from aws.osml.models.server_utils import detect_to_feature, setup_server

# enable exceptions for GDAL
gdal.UseExceptions()

# set up our flask app
app = Flask(__name__)
app.logger.setLevel(logging.ERROR)

# Optional ENV configurations
BBOX_PERCENTAGE = float(os.environ.get("BBOX_PERCENTAGE", 0.1))
ENABLE_SEGMENTATION = os.environ.get("ENABLE_SEGMENTATION", "False").lower() == "true"


def gen_center_bbox(width: int, height: int, bbox_percentage: float) -> List[float]:
    """
    Create a single detection bbox that is at the center of and sized proportionally to the image
    :param bbox_percentage: the size of the bounding box and poly, relative to the image, to return
    :param width: Raster width of the image passed in
    :param height: Raster height of the image passed in
    :return:bbox: Segmented bbox array for center detection
    """
    center_xy = width / 2, height / 2
    fixed_object_size_xy = width * bbox_percentage, height * bbox_percentage
    return [
        center_xy[0] - fixed_object_size_xy[0],
        center_xy[1] - fixed_object_size_xy[1],
        center_xy[0] + fixed_object_size_xy[0],
        center_xy[1] + fixed_object_size_xy[1],
    ]


def gen_center_polygon(width: int, height: int, bbox_percentage: float) -> List[List[float]]:
    center = 0, 0
    center_xy = width / 2, height / 2
    radius = bbox_percentage
    # 20 is a nice circle, 3 is a triangle, etc
    number_of_vertices = 6
    circle = CirclePolygon(center, radius, resolution=number_of_vertices)
    poly_path = circle.get_path().vertices.tolist()
    # This is part of CV model requirements, to have (only) closed polygons
    poly_path.append(poly_path[0])
    # This moves poly to nonzero 0-1 coords
    nonzero_circle = [((x + 1) / 2, (y + 1) / 2) for (x, y) in poly_path]
    # Project to the correct percentage of our image coordinates
    scale = [bbox_percentage * width, bbox_percentage * height]
    # Do final scaling of coordinates, and w/h translation to ensure within bounds of the bbox
    center_polygon = [
        [round(x * scale[0] + center_xy[0], 4), round(y * scale[1] + center_xy[1], 4)] for (x, y) in nonzero_circle
    ]
    return center_polygon


def gen_center_detect(width: int, height: int, bbox_percentage: float) -> Dict:
    """
    Create  circular polygon that is at the center of and sized proportionally to the bbox
    :param bbox_percentage: the size of the bounding box and poly, relative to the image, to return
    :param width: Raster width of the image passed in
    :param height: Raster height of the image passed in
    :return: geojson: Segmented polygon for center detection
    this draws a polygon with the same center
    and width and height percentage polygon can be a circle, or a hexagon, or triangle, etc. - based on the
    number_of_vertices there is a chance this is not centered as we'd like - meanwhile this will work as-is for initial
     OSML segmentation 'passthrough'

    """
    center_polygon = None
    if ENABLE_SEGMENTATION:
        center_polygon = gen_center_polygon(width, height, bbox_percentage)

    geojson_feature = detect_to_feature(gen_center_bbox(width, height, bbox_percentage), center_polygon)
    return {"type": "FeatureCollection", "features": [geojson_feature]}


@app.route("/ping", methods=["GET"])
def healthcheck() -> Response:
    """
    This is a health check that will always pass since this is a stub model.

    :return: Response: Status code (200) indicates all is well
    """
    app.logger.debug("Responding to health check")
    return Response(response="\n", status=200)


@app.route("/invocations", methods=["POST"])
def predict() -> Response:
    """
    This is the model invocation endpoint for the model container's REST
    API. The binary payload, in this case an image, is taken from the request
    parsed to ensure it is a valid image. This is a stub implementation that
    will always return the fixed set of detections for a valid input image.

    :return: Response: Contains the GeoJSON results or an error status
    """
    app.logger.debug("Invoking centerpoint model endpoint")
    temp_ds_name = "/vsimem/" + token_hex(16)
    gdal_dataset = None
    try:
        # Load the file from the request memory buffer
        gdal.FileFromMemBuffer(temp_ds_name, request.get_data())
        try:
            gdal_dataset = gdal.Open(temp_ds_name)
        # If it failed to load return the failed Response
        except RuntimeError:
            return Response(response="Unable to parse image from request!", status=400)

        geojson_feature_collection = gen_center_detect(gdal_dataset.RasterXSize, gdal_dataset.RasterYSize, BBOX_PERCENTAGE)
        app.logger.debug(json.dumps(geojson_feature_collection))
        # Send back the detections
        return Response(response=json.dumps(geojson_feature_collection), status=200)
    except Exception as err:
        app.logger.warning("Image could not be processed by the centerpoint model server.", exc_info=True)
        app.logger.warning(err)
        return Response(response="Unable to process request.", status=500)
    finally:
        if gdal_dataset is not None:
            if temp_ds_name is not None:
                gdal.Unlink(temp_ds_name)
            del gdal_dataset


# pragma: no cover
if __name__ == "__main__":
    setup_server(app)



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

# Telling flake8 to not flag errors in this file. It is normal that these classes are imported but not used in an
# __init__.py file.
# flake8: noqa

from .server_utils import detect_to_feature, setup_server



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

import argparse
import logging
from secrets import token_hex
from typing import Dict, List, Optional, Union

from flask import Flask
from osgeo import gdal

# Enable exceptions for GDAL
gdal.UseExceptions()


def setup_server(app: Flask):
    """
    The assumption is that this script will be the ENTRYPOINT for the inference
    container. SageMaker will launch the container with the "serve" argument. We
    also have the option of using multiple models from this single container;
    only one model will be active at a time (i.e., this is not a Multi Model Server),
    so it can be selected by name using the "model" parameter.

    :param app: The flask application to set up
    :return: None
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("command", nargs="?", default="serve")
    parser.add_argument("-v", "--verbose", default=False)
    args = parser.parse_args()

    # Set up the logging for the Flask application
    # and log the startup information
    configure_logging(args.verbose)
    logging.info("Initializing REST Model Server...")
    for arg, value in sorted(vars(args).items()):
        logging.info("Argument %s: %r", arg, value)

    # Start the simple web application server using Waitress.
    # Flask's app.run() is only intended to be used in development
    #  mode, so this provides a solution for hosting the application.
    from waitress import serve

    serve(app, host="0.0.0.0", port=8080, clear_untrusted_proxy_headers=True)


def configure_logging(verbose: bool = False) -> None:
    """
    Configure application logging. Note the timestamp for the
    log record is available in CloudWatch if this is used in a
    non-AWS environment, you can add %(asctime)s to the start of
    the format string.

    :param verbose: True, if the DEBUG log level should be used, defaults to False
    :return: None
    """
    logging_level = logging.INFO
    if verbose:
        logging_level = logging.DEBUG

    logging.basicConfig(level=logging_level, format="%(levelname)-8s : %(message)s")


def detect_to_feature(
    fixed_object_bbox: List[float],
    fixed_object_mask: Optional[List[List[float]]] = None,
    detection_score: Optional[float] = 1.0,
    detection_type: Optional[str] = "sample_object",
) -> Dict[str, Union[str, list]]:
    """
    Converts the bbox object into a sample GeoJSON formatted detection.
    The model container does not normally provide the world coordinates,
    so they're defaulted to 0,0 here since GeoJSON features require a geometry.

    :param detection_type: Class assigned to the detection
    :param detection_score: Confidence score assigned to the detection
    :param fixed_object_mask: Polygon version of mask generated by detectron2
    :param fixed_object_bbox: Bounding box to transform into a geojson feature
    :return: dict: Dictionary representation of a geojson feature
    """
    feature = {
        "type": "Feature",
        "geometry": {"coordinates": [0.0, 0.0], "type": "Point"},
        "id": token_hex(16),
        "properties": {
            "bounds_imcoords": fixed_object_bbox,
            "detection_score": detection_score,
            "feature_types": {detection_type: detection_score},
            "image_id": token_hex(16),
        },
    }

    if fixed_object_mask is not None:
        feature["properties"]["geom_imcoords"] = fixed_object_mask

    return feature



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

import json
import logging
import os
import uuid
import warnings
from typing import Dict, List, Optional, Union

import cv2
import numpy as np
import torch
from detectron2 import model_zoo
from detectron2.config import get_cfg
from detectron2.engine import DefaultPredictor
from detectron2.structures.instances import Instances
from flask import Flask, Request, Response, request
from osgeo import gdal

from aws.osml.models import setup_server

ENABLE_SEGMENTATION = os.environ.get("ENABLE_SEGMENTATION", "False").lower() == "true"

# enable exceptions for GDAL
gdal.UseExceptions()

app = Flask(__name__)
app.logger.setLevel(logging.ERROR)


def build_predictor() -> DefaultPredictor:
    """
    Create a single detection predictor to detect aircraft
    :return: DefaultPredictor
    """
    # Load the prebuilt plane model w/ Detectron2
    cfg = get_cfg()
    # If we can't find a gpu
    if not torch.cuda.is_available():
        cfg.MODEL.DEVICE = "cpu"
        app.logger.warning("GPU not found, running in CPU mode!")
    # Set to only expect one class (aircraft)
    cfg.MODEL.ROI_HEADS.NUM_CLASSES = 1
    # Set the detection threshold to 90%
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.9
    # Add project-specific config used for training to remove warnings
    cfg.merge_from_file(model_zoo.get_config_file("COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml"))
    # Path to the model weights
    cfg.MODEL.WEIGHTS = os.getenv(
        os.path.join("MODEL_WEIGHTS"), os.path.join("/home/osml-models/assets/", "model_weights.pth")
    )

    # Build the detectron2 default predictor
    return DefaultPredictor(cfg)


def mask_to_polygon(mask: torch.Tensor) -> List[List[float]]:
    """
    Convert a detectron2 instance mask in the form of a Tensor to a list-form polygon representing the mask.

    :param mask: A detectron2 instance mask in the form of a Tensor
    :return: A list form polygon representing the mask
    """
    # Convert tensor to numpy
    mask_np = mask.cpu().numpy().astype(np.uint8)

    # Convert mask to 255 (fully white) and 0 (black)
    _, binary_mask = cv2.threshold(mask_np, 0.5, 255, cv2.THRESH_BINARY)

    # Find contours
    contours, _ = cv2.findContours(binary_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)

    # Simplify contour if you want to save some cost in exchange
    # for reduced resolution on the masks
    # epsilon = 0.005 * cv2.arcLength(contours[0], True)
    # approx = cv2.approxPolyDP(contours[0], epsilon, True)

    # Convert contour to a list of lists
    polygon = [point[0].tolist() for point in contours[0]]

    return polygon


def instances_to_feature_collection(
    instances: Instances, image_id: Optional[str] = str(uuid.uuid4())
) -> Dict[str, Union[str, list]]:
    """
    Convert the gRPC response from the GetDetection call into a GeoJSON output.
    Each detection is a feature in the collection, including image coordinates,
    score, and type identifier as feature properties.

    :param instances: Detectron2 result instances
    :param image_id: Identifier for the processed image (optional)
    :return: FeatureCollection object containing detections
    """
    geojson_feature_collection_dict = {"type": "FeatureCollection", "features": []}
    if instances:
        # Get the bounding boxes for this image
        bboxes = instances.pred_boxes.tensor.cpu().numpy().tolist()

        # Get the scores for this image
        scores = instances.scores.cpu().numpy().tolist()

        # Default masks to None
        masks = None

        # Get the polygon masks for this image if segmentation is enabled
        if ENABLE_SEGMENTATION:
            masks = instances.pred_masks.cpu()

        for i in range(0, len(bboxes)):
            feature = {
                "type": "Feature",
                "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
                "id": str(uuid.uuid4()),
                "properties": {
                    "bounds_imcoords": bboxes[i],
                    "detection_score": float(scores[i]),
                    "feature_types": {"aircraft": float(scores[i])},
                    "image_id": image_id,
                },
            }
            if masks is not None:
                feature["properties"]["geom_imcoords"] = mask_to_polygon(masks[i])
            app.logger.debug(feature)
            geojson_feature_collection_dict["features"].append(feature)
    else:
        app.logger.debug("No features found!")

    return geojson_feature_collection_dict


def request_to_instances(req: Request) -> Union[Instances, None]:
    """
    Use GDAL to open the image. The binary payload from the HTTP request is used to
    create an in-memory VFS for GDAL which is then opened to decode the image into
    a dataset which will give us access to a NumPy array for the pixels. Then
    use that image to create detectron2 detection instances.

    :param req: Request: the flask request object passed into the SM endpoint
    :return: Either a set of detectron2 detection instances or nothing
    """
    # Set up default variables
    temp_ds_name = "/vsimem/" + str(uuid.uuid4())
    gdal_dataset = None
    instances = None
    try:
        # Load the binary memory buffer sent to the model
        gdal.FileFromMemBuffer(temp_ds_name, req.get_data())
        gdal_dataset = gdal.Open(temp_ds_name)

        # Read GDAL dataset and convert to a numpy array
        image_array = gdal_dataset.ReadAsArray()

        # Handling of different image shapes
        if image_array.ndim == 2:  # For grayscale images without a channel dimension
            # Reshape to add a channel dimension and replicate across 3 channels for RGB
            image_array = np.stack([image_array] * 3, axis=0)
        elif image_array.shape[0] == 1:  # For grayscale images with a channel dimension
            # Replicate the single channel across 3 channels for RGB
            image_array = np.repeat(image_array, 3, axis=0)
        elif image_array.shape[0] == 4:  # For images with an alpha channel
            # Remove the alpha channel
            image_array = image_array[:3, :, :]

        # Conversion to uint8 (ensure this is done after ensuring 3 channels)
        image_array = (image_array * 255).astype(np.uint8)

        # Transpose the array from (channels, height, width) to (height, width, channels)
        image = np.transpose(image_array, (1, 2, 0))

        # grab detections from Detectron2
        app.logger.debug(f"Running D2 on image array: {image}")

        # PyTorch can often give warnings about upcoming changes
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            instances = aircraft_predictor(image)["instances"]
    except Exception as err:
        app.logger.warning(f"Unable to load image from request: {err}")
        return None
    finally:
        try:
            if gdal_dataset is not None:
                if temp_ds_name is not None:
                    gdal.Unlink(temp_ds_name)
                del gdal_dataset
        except Exception as err:
            app.logger.warning(f"Unable to cleanup gdal dataset: {err}")

    return instances


# Build our aircraft predictor
aircraft_predictor = build_predictor()


@app.route("/ping", methods=["GET"])
def healthcheck() -> Response:
    """
    This is a health check that will always pass since this is a stub model.

    :return: Successful status code (200) indicates all is well
    """
    app.logger.debug("Responding to health check")
    return Response(response="\n", status=200)


@app.route("/invocations", methods=["POST"])
def predict() -> Response:
    """
    This is the model invocation endpoint for the model container's REST
    API. The binary payload, in this case an image, is taken from the request
    parsed to ensure it is a valid image. This is a stub implementation that
    will always return a fixed set of detections for a valid input image.

    :return: Response: Contains the GeoJSON results or an error status
    """
    app.logger.debug("Invoking model endpoint using the Detectron2 Aircraft Model!")
    try:
        # Load the image into memory and get detection instances
        app.logger.debug("Loading image request.")
        instances = request_to_instances(request)

        # Generate a geojson feature collection that we can return
        geojson_detects = instances_to_feature_collection(instances)
        app.logger.debug(f"Sending geojson to requester: {json.dumps(geojson_detects)}")

        # Send back the detections
        return Response(response=json.dumps(geojson_detects), status=200)
    except Exception as err:
        app.logger.debug(err)
        return Response(response="Unable to process request!", status=500)


# pragma: no cover
if __name__ == "__main__":
    setup_server(app)



#  Copyright 2023-2024 Amazon.com, Inc. or its affiliates.

import json
import logging
import math
import os
import random
from random import randrange
from secrets import token_hex
from typing import Dict, Union

from flask import Flask, Response, request
from osgeo import gdal

from aws.osml.models import detect_to_feature
from aws.osml.models.server_utils import setup_server

app = Flask(__name__)
app.logger.setLevel(logging.ERROR)

# Optional ENV configurations
BBOX_PERCENTAGE = float(os.environ.get("BBOX_PERCENTAGE", 0.1))
FLOOD_VOLUME = int(os.environ.get("FLOOD_VOLUME", 100))
ENABLE_SEGMENTATION = os.environ.get("ENABLE_SEGMENTATION", "False").lower() == "true"


def gen_flood_detects(height: int, width: int, bbox_percentage: float) -> Dict[str, Union[str, list]]:
    """
    Generate a random detection within the input image given a buffer percentage that
    limits the bounding boxes we generate to always fall within the image bounds.

    :param bbox_percentage: The size of the bounding box to produce.
    :param width: Width of the image tile.
    :param height: Height of the image tile.
    :return: Union[gdal.Dataset, None]: either the gdal dataset or nothing
    """
    geojson_features = []
    for i in range(FLOOD_VOLUME):
        fixed_object_size_xy = math.ceil(width * bbox_percentage), math.ceil(height * bbox_percentage)
        gen_x = randrange(fixed_object_size_xy[0], width - fixed_object_size_xy[0])
        gen_y = randrange(fixed_object_size_xy[1], height - fixed_object_size_xy[1])
        fixed_object_bbox = [
            gen_x - fixed_object_size_xy[0],
            gen_y - fixed_object_size_xy[1],
            gen_x + fixed_object_size_xy[0],
            gen_y + fixed_object_size_xy[1],
        ]
        fixed_object_mask = None
        if ENABLE_SEGMENTATION:
            fixed_object_mask = [
                [gen_x - fixed_object_size_xy[0], gen_y + fixed_object_size_xy[1]],
                [gen_y - fixed_object_size_xy[0], gen_x + fixed_object_size_xy[0]],
                [gen_x + fixed_object_size_xy[0], gen_y + fixed_object_size_xy[1]],
                [gen_y + fixed_object_size_xy[1], gen_x + fixed_object_size_xy[0]],
                [gen_x - fixed_object_size_xy[0], gen_y + fixed_object_size_xy[1]],
            ]
        # Create a feature with a random confidence score for each random detect
        feature = detect_to_feature(fixed_object_bbox, fixed_object_mask, random.uniform(0, 1))
        geojson_features.append(feature)

    geojson_feature_collection_dict = {"type": "FeatureCollection", "features": geojson_features}

    return geojson_feature_collection_dict


@app.route("/ping", methods=["GET"])
def healthcheck() -> Response:
    """
    This is a health check that will always pass since this is a stub model.

    :return: A successful status code (200) indicates all is well
    """
    app.logger.debug("Responding to health check")
    return Response(response="\n", status=200)


@app.route("/invocations", methods=["POST"])
def predict() -> Response:
    """
    This is the model invocation endpoint for the model container's REST
    API. The binary payload, in this case an image, is taken from the request
    parsed to ensure it is a valid image. This is a stub implementation that
    will always return a fixed set of detections for a valid input image.

    :return: Response: Contains the GeoJSON results or an error status
    """
    app.logger.debug("Invoking flood model endpoint!")
    temp_ds_name = "/vsimem/" + token_hex(16)
    gdal_dataset = None
    try:
        # load the file from the request memory buffer
        gdal.FileFromMemBuffer(temp_ds_name, request.get_data())
        try:
            gdal_dataset = gdal.Open(temp_ds_name)

        # if it failed to load return the failed Response
        except RuntimeError:
            return Response(response="Unable to parse image from request!", status=400)

        # generate random flood detections
        geojson_detects = gen_flood_detects(gdal_dataset.RasterXSize, gdal_dataset.RasterYSize, BBOX_PERCENTAGE)

        # send back the detections
        return Response(response=json.dumps(geojson_detects), status=200)

    except Exception as err:
        app.logger.warning("Image could not be processed by the test model server.", exc_info=True)
        app.logger.warning(err)
        return Response(response="Unable to process request.", status=500)

    finally:
        if gdal_dataset is not None:
            if temp_ds_name is not None:
                gdal.Unlink(temp_ds_name)
            del gdal_dataset


if __name__ == "__main__":  # pragma: no cover
    setup_server(app)
